# Copyright 2025 Georges Racinet <georges.racinet@cloudcrane.io>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Management of Project and Group Mercurial configuration files.
"""
import os
from pathlib import Path

INITIAL_PROJECT_HGRC_HEADER = b'\n'.join((
    b"# This is the specific configuration for this repository",
    b"# You may modify it on the file system if you follow the explanations",
    b"# in comments below",
    b"",
))

INITIAL_PROJECT_GROUP_INCLUSION_HEADER = b'\n'.join((
    b"#",
    b"# By default, it includes the configuration from the enclosing group",
    b"# or user namespace, but all the settings defined in the present file",
    b"# will take precedence",
    b"# as long as they are below the %include line",
    b"# You may also remove the %include line altogether, although it is",
    b"# cleaner to use the `PUT projects/:id/hgrc` API endpoint with",
    b'# `"inherit": false`` for this purpose. ',
    b"",
    b"# Group-level configuration:",
    b"",
))

INITIAL_PROJECT_HGRC_MANAGED_INCLUSION = b'\n'.join((
    b"# specific configuration managed by Heptapod. "
    b"# Do not remove it, it can come back at any time.",
    b"%include hgrc.managed",
    b"",
))

INITIAL_PROJECT_HGRC_MANAGED = b'\n'.join((
    b"# This file is entirely managed by Heptapod\n",
    b"# Any modification you might do can disappear at any time\n"
))


INITIAL_SUBGROUP_HGRC = b'\n'.join((
    b"# This is the Mercurial configuration for this subgroup.",
    b"# You may modify it freely at your will.",
    b"#",
    b"# By default, it includes the configuration from the enclosing",
    b"# group, but all the settings defined in the present file will",
    b"# take precedence as long as they are below the %include line",
    b"# You may also remove the %include line altogether if you prefer.",
    b"",
    b"# enclosing group configuration",
    b"%include ../hgrc",
    b"",
))

NO_PARENT = Path('.')


def init_project_hgrc_files(repo, repo_rel_path, project_full_path):
    """Create all needed configuration files for the given project.

    :param repo_rel_path: the relative path from storage root of the
        repository, as given in the `relative_path` field of the Gitaly
        ``Repository`` message
    :param project_full_path: the full Project applicative path,
        as given in the `gl_project_path` field of the Gitaly ``Repository``
        message.
    """
    split = project_full_path.rsplit('/', 1)
    if len(split) == 1:
        group_rpath = None
    else:
        group_rpath = Path(split[0])
        hgrc_rpath = Path(repo_rel_path) / '.hg'
        # Path.relative_to gains the `walk_up` option in Python 3.12
        group_rel_repo_path = os.path.relpath(group_rpath, hgrc_rpath)
        group_include_line = "%include {}\n".format(
            Path(group_rel_repo_path) / 'hgrc')

    orig_umask = os.umask(0o077)
    try:
        with repo.wlock():
            with repo.vfs(b'hgrc',
                          mode=b'wb',
                          atomictemp=True,
                          checkambig=True) as fobj:
                fobj.write(INITIAL_PROJECT_HGRC_HEADER)
                if group_rpath is not None:
                    fobj.write(INITIAL_PROJECT_GROUP_INCLUSION_HEADER)
                    fobj.write(os.fsencode(group_include_line))
                fobj.write(INITIAL_PROJECT_HGRC_MANAGED_INCLUSION)

            with repo.vfs(b'hgrc.managed',
                          mode=b'wb',
                          atomictemp=True,
                          checkambig=True) as fobj:
                fobj.write(INITIAL_PROJECT_HGRC_MANAGED)

            if group_rpath is None:
                return

            repo_path = Path(os.fsdecode(repo.path))
            group_abspath = (repo_path / group_rel_repo_path).resolve()
            os.makedirs(group_abspath, exist_ok=True)
            parent_group = group_rpath.parent
            while parent_group != NO_PARENT:
                group_hgrc = (group_abspath / 'hgrc')
                if not group_hgrc.exists():
                    group_hgrc.write_bytes(INITIAL_SUBGROUP_HGRC)
                group_rpath = parent_group
                parent_group = group_rpath.parent
                group_abspath = group_abspath.parent
    finally:
        os.umask(orig_umask)



import boto3
import os
from urllib.parse import urlunparse
from mercurial import util as hgutil


BUCKET_URL_PREFIX = b'bucket:/'
"""Using a special scheme meant for objects in private buckets.

We do not use simply `s3://` because in this case the next segment
is usualy the bucket.
"""


def s3_client_bucket(repo, public=False):

    def s3_conf(subkey, boolean=False):
        get = repo.ui.configbool if boolean else repo.ui.config
        value = get(b'heptapod',
                    b'clone-bundles.s3.' + subkey.encode('ascii'))
        if boolean:
            return value
        elif value is not None:
            return value.decode('ascii')

    bucket_key = 'public-bucket' if public else 'private-bucket'
    scheme = 'https' if s3_conf('tls', boolean=True) else 'http'
    endpoint_url = urlunparse((scheme, s3_conf('endpoint'), '', '', '', ''))
    session = boto3.session.Session()
    ca_bundle = repo.ui.config(b'web', b'cacerts')
    if ca_bundle:
        os.environ['AWS_CA_BUNDLE'] = os.fsdecode(
            hgutil.expandpath(ca_bundle)
        )
    client = session.client(
        service_name='s3',
        region_name=s3_conf('region'),
        endpoint_url=endpoint_url,
        aws_access_key_id=s3_conf('access_key'),
        aws_secret_access_key=s3_conf('secret_key'),
    )
    return client, s3_conf(bucket_key)



# Copyright 2025 Georges Racinet <georges.racinet@cloudcrane.io>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import stat

from ..testhelpers import RepoWrapper
from ..hgrc import (
    init_project_hgrc_files,
)

PRIVATE_DIR = 0o700
PRIVATE_FILE = 0o600


def hgrc_lines(wrapper, ext=''):
    return [line.strip()
            for line in (wrapper.path / f'.hg/hgrc{ext}').readlines()]


def perms(path):
    # ignore bits that can be due to the path being in /tmp etc.
    return stat.S_IMODE(path.stat().mode)


def test_project_hgrc_files(tmpdir):
    gl_rpath = '@hashed/12/34/looongsha.hg'
    wrapper = RepoWrapper.init(tmpdir / gl_rpath)

    init_project_hgrc_files(wrapper.repo, gl_rpath, 'proj-no-group')
    includes = [line for line in hgrc_lines(wrapper)
                if line.startswith('%include')]
    assert includes == ['%include hgrc.managed']

    init_project_hgrc_files(wrapper.repo, gl_rpath, 'group/proj')
    lines = hgrc_lines(wrapper)
    assert '%include ../../../../../group/hgrc' in lines
    assert '%include hgrc.managed' in lines
    assert 'managed by Heptapod' in hgrc_lines(wrapper, ext='.managed')[0]
    group_path = tmpdir / 'group'
    assert group_path.exists()
    assert perms(group_path) == PRIVATE_DIR
    assert not (group_path / 'hgrc').exists()

    init_project_hgrc_files(wrapper.repo, gl_rpath, 'group/sub/proj')
    lines = hgrc_lines(wrapper)
    assert '%include ../../../../../group/sub/hgrc' in lines
    assert '%include hgrc.managed' in lines
    subgroup_path = group_path / 'sub'
    assert subgroup_path.exists()
    assert perms(subgroup_path) == PRIVATE_DIR
    assert not (group_path / 'hgrc').exists()
    subgroup_hgrc_path = (subgroup_path / 'hgrc')
    assert subgroup_hgrc_path.exists()
    assert perms(subgroup_hgrc_path) == PRIVATE_FILE
    assert '%include ../hgrc' in subgroup_hgrc_path.read()

    # Creating a new project in the same subgroup does not overwrite
    # existing group hgrc files
    subgroup_hgrc_path.write_binary(b"# just a comment, no inclusion")
    gl_rpath2 = '@hashed/56/78/looongsha.hg'
    wrapper = RepoWrapper.init(tmpdir / gl_rpath2)
    init_project_hgrc_files(wrapper.repo, gl_rpath2, 'group/sub/proj2')
    # Main assertion
    assert 'just a comment' in subgroup_hgrc_path.read()
    # Remaining general assertions for sanity
    lines = hgrc_lines(wrapper)
    assert '%include ../../../../../group/sub/hgrc' in lines
    assert '%include hgrc.managed' in lines
    assert perms(subgroup_path) == PRIVATE_DIR
    assert not (group_path / 'hgrc').exists()
    assert perms(subgroup_hgrc_path) == PRIVATE_FILE
